/*
 * Copyright 2011 Hanson Robokind LLC.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.robokind.api.animation.utils;

import java.util.List;
import java.util.Map;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.osgi.framework.BundleContext;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceReference;
import org.robokind.api.animation.Animation;
import org.robokind.api.animation.player.AnimationJob;
import org.robokind.api.animation.player.AnimationPlayer;
import org.robokind.api.common.osgi.ClassTracker;
import org.robokind.api.common.osgi.OSGiUtils;

/**
 *
 * @author Matthew Stevenson <www.robokind.org>
 */
public class AnimationUtils {
    private final static Logger theLogger = Logger.getLogger(AnimationUtils.class.getName());
    private static ClassTracker<ChannelsParameterSource> theChannelParamTracker;
    private static ClassTracker<AnimationEditListener> theListeners;
    
    private static ClassTracker<ChannelsParameterSource> getChanParamTracker(){
        if(theChannelParamTracker != null){
            return theChannelParamTracker;
        }
        BundleContext context = OSGiUtils.getBundleContext(ChannelsParameterSource.class);
        if(context == null){
            return null;
        }
        theChannelParamTracker = new ClassTracker<ChannelsParameterSource>(
                context, ChannelsParameterSource.class.getName(), null, null);
        return theChannelParamTracker;
    }
    
    public static ChannelsParameterSource getChannelsParameterSource(){
       ClassTracker<ChannelsParameterSource> tracker = getChanParamTracker();
       return tracker == null ? null : tracker.getTopService();
    }
    
    public static void writePositions(long x, Map<Integer,Double> channelPositions){
        List<AnimationEditListener> listeners = getListeners();
        if(listeners == null){
            return;
        }
        for(AnimationEditListener listener : listeners){
            listener.handlePositions(x, channelPositions);
        }
    }
    
    private static List<AnimationEditListener> getListeners(){
        ClassTracker<AnimationEditListener> tracker = getAELTracker();
        return tracker == null ? null : tracker.getServices();
    }
    
    private static ClassTracker<AnimationEditListener> getAELTracker(){
        if(theListeners != null){
            return theListeners;
        }
        BundleContext context = OSGiUtils.getBundleContext(AnimationEditListener.class);
        if(context == null){
            return null;
        }
        theListeners = new ClassTracker<AnimationEditListener>(
                context, AnimationEditListener.class.getName(), null, null);
        return theListeners;
    }
    /**
     * Returns a ServiceReference for an AnimationPlayer matching the filter if
     * it is not null.
     * @param context BundleContext to use
     * @param filter option OSGi filter String
     * @return ServiceReference for an AnimationPlayer matching the filter if
     * it is not null
     */
    public static ServiceReference getAnimationPlayerReference(
            BundleContext context, String filter){
        if(context == null){
            throw new NullPointerException();
        }
        ServiceReference[] refs = null;
        try{
            refs= context.getServiceReferences(AnimationPlayer.class.getName(), filter);
        }catch(InvalidSyntaxException ex){
            theLogger.log(Level.WARNING, "Invlaid filter: " + filter + 
                    ".  Could not fetch AnimationPlayer.", ex);
        }
        if(refs == null){
            theLogger.log(Level.WARNING, "Could not find AnimationPlayer");
            return null;
        }
        return refs[0];
    }
    /**
     * Fetches an AnimationPlayer and plays the given Animation.  If a filter
     * String is provided, it is used to match an AnimationPlayer.
     * @param context BundleContext to use
     * @param filter option OSGi filter String
     * @param anim Animation to play
     * @return AnimationJob created from playing the Animation, returns null
     * if unsuccessful
     */
    public static AnimationJob playAnimation(
            BundleContext context, String filter, Animation anim){
        return _playAnimation(context, filter, anim, null, null);
    }
    /**
     * Fetches an AnimationPlayer and plays the given Animation with the given
     * start and stop time.  If a filter
     * String is provided, it is used to match an AnimationPlayer.
     * @param context BundleContext to use
     * @param filter option OSGi filter String
     * @param anim Animation to play
     * @param startTime Animation start time in milliseconds from the beginning 
     * of the animation
     * @param stopTime Animation stop time in milliseconds from the beginning 
     * of the animation
     * @return AnimationJob created from playing the Animation, returns null
     * if unsuccessful
     */
    public static AnimationJob playAnimation(
            BundleContext context, String filter, Animation anim, 
            long startTime, long stopTime){
        return _playAnimation(context, filter, anim, startTime, stopTime);
    }
    
    private static AnimationJob _playAnimation(
            BundleContext context, String filter, Animation anim, 
            Long startTime, Long stopTime){
        if(context == null || anim == null){
            throw new NullPointerException();
        }
        ServiceReference ref = 
                AnimationUtils.getAnimationPlayerReference(context, filter);
        if(ref == null){
            return null;
        }
        Object obj = context.getService(ref);
        if(obj == null || !(obj instanceof AnimationPlayer)){
            context.ungetService(ref);
            return null;
        }
        AnimationPlayer player = (AnimationPlayer)obj;
        AnimationJob job = null;
        if(player != null){
            if(startTime != null && stopTime != null){
                job = player.playAnimation(anim, startTime, stopTime);
            }else{
                job = player.playAnimation(anim);
            }
        }
        context.ungetService(ref);
        return job;
    }
}
