/*
 * Copyright 2011 Hanson Robokind LLC.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.robokind.api.common.osgi;

import java.util.Map.Entry;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.FrameworkUtil;
import org.osgi.framework.InvalidSyntaxException;
import org.osgi.framework.ServiceReference;

/**
 * Utility methods for working with OSGi.
 * @author Matthew Stevenson <www.robokind.org>
 */
public class OSGiUtils {
    private final static Logger theLogger = Logger.getLogger(OSGiUtils.class.getName());
    /**
     * Returns the BundleContext associated with the given Class.
     * @param clazz Class associated with a Bundle
     * @return the BundleContext associated with the given Class, if no Bundle
     * is found for the Class null is returned
     */
    public static BundleContext getBundleContext(Class clazz){
        Bundle bundle = FrameworkUtil.getBundle(clazz);
        if(bundle == null){
            return null;
        }
        return bundle.getBundleContext();     
    }
    
    /**
     * Creates a filter string which matches all of the properties given.
     * @param props Properties to match
     * @return OSGi filter string, null if props is null or empty
     */
    public static String createServiceFilter(Properties props){
        if(props == null || props.isEmpty()){
            return null;
        }
        String filter = "";
        for(Entry e : props.entrySet()){
            String key = (String)e.getKey();
            String val = (String)e.getValue();
            filter = String.format("%s(%s=%s)", filter,key,val);
        }
        return String.format("(&%s)", filter);
    }
    
    /**
     * Creates a filter string which matches the key and value given
     * @param key key
     * @param val value
     * @return OSGi filter string, null if props is null or empty
     */
    public static String createServiceFilter(String key, String val){
        if(key == null || val == null){
            return null;
        }
        return String.format("(%s=%s)",key,val);
    }
    
    /**
     * Checks the OSGi Service Registry for a Service with the given class and
     * matching the given filter. Returns true if a matching ServiceReference is
     * found, return null if there is an InvalidSyntaxException.
     * @param context BundleContext to use
     * @param clazz Class name to match
     * @param filter OSGi Service filter String
     * @return true if a matching ServiceReference is found, returns null if 
     * there is an InvalidSyntaxException
     */
    public static Boolean serviceExists(
            BundleContext context, String clazz, String filter){
        try{
            ServiceReference[] refs = 
                    context.getAllServiceReferences(clazz, filter); 
            boolean exists = (refs != null);
            return exists;
        }catch(InvalidSyntaxException ex){
            theLogger.log(Level.WARNING, 
                    "Invalid OSGi filter syntax String: " + filter + ".", ex);
            return null;
        }
    }
    
    /**
     * Checks the OSGi Service Registry for an existing Service matching the
     * given filter String and one or more of the given Class names.
     * @param context BundleContext to use
     * @param classes one or more of the classes must be matched
     * @param filter OSGi Service filter String to match
     * @return true if a ServiceReference is found matching one or more of the
     * given classes and the filter String, return null if there is an 
     * InvalidSyntaxException
     */
    public static Boolean serviceExists(
            BundleContext context, Class[] classes, String filter) {
        for(Class c : classes){
            Boolean b = OSGiUtils.serviceExists(context, c.getName(), filter);
            if(!b){
                continue;
            }else if(b){
                theLogger.log(Level.WARNING, 
                        "Found existing Robot Service: {0}, matching {1}.", 
                        new Object[]{c, filter});
                return true;
            }
            //InvalidSyntaxException, we cannot be sure if a service exists
            return null; 
        }
        return false;
    }
}
