/*
 * Copyright 2011 Hanson Robokind LLC.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.robokind.nbui.motion.actions;

import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.Map;
import java.util.Properties;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.swing.JFileChooser;
import org.osgi.framework.BundleContext;
import org.osgi.framework.ServiceRegistration;
import org.robokind.api.motion.Robot;
import org.robokind.api.motion.servos.ServoRobot;
import org.robokind.api.motion.servos.ServoController;
import org.robokind.impl.motion.config.RobotConfigXMLReader;
import org.robokind.api.common.osgi.OSGiUtils;
import org.robokind.api.common.services.ServiceConnectionDirectory;
import org.robokind.api.motion.utils.JointGroup;
import org.robokind.api.motion.utils.RobotUtils;
import org.robokind.extern.utils.apache_commons_configuration.ConfigUtils;
import org.robokind.impl.motion.utils.RobotJointGroup;
import org.robokind.impl.motion.utils.RobotJointGroupConfigXMLReader;

/**
 * 
 * @author Matthew Stevenson <www.robokind.org>
 */
public final class ConnectAction implements ActionListener {
    private final static Logger theLogger = 
            Logger.getLogger(ConnectAction.class.getName());
    
    private final static long theDefaultBlenderInterval = 40L;

    @Override
    public void actionPerformed(ActionEvent e) {
        //File file = chooseFile();
        boolean jointGroup = false;
        File file = predfinedFile("./resources/robot.xml");
        if(file == null){
            file = chooseFile();
            if(file == null){
                return;
            }
        }else{
            jointGroup = true;
        }
        BundleContext context = OSGiUtils.getBundleContext(Robot.class);
        if(context == null){
            theLogger.warning(
                    "Unable to load Robot.  Could not find BundleContext.");
            return;
        }
        Robot robot = ServiceConnectionDirectory.buildService(context, 
                ServoRobot.VERSION, RobotConfigXMLReader.VERSION, 
                file, File.class, Robot.class);
        if(robot == null){
            theLogger.warning(
                    "Unable to load Robot.  Could not lind loaders.");
            return;
        }
        robot.connect();
        ServiceRegistration reg = 
                RobotUtils.registerRobot(context, robot, null);
        if(reg == null){
            theLogger.log(Level.INFO, "Failed registering Robot service.");
            return;
        }
        theLogger.log(Level.INFO, "Robot Service Registered.");
        if(jointGroup){
            loadJointGroup(context);
        }
        if(robot instanceof ServoRobot){
            setServoRobot(context, (ServoRobot)robot);
        }
        RobotUtils.startDefaultBlender(
                context, robot.getRobotId(), theDefaultBlenderInterval);
    }
    
    private void setServoRobot(BundleContext context, ServoRobot robot){
        ServoRobot sr = (ServoRobot)robot;
        Map<ServoController.Id,ServoController> controllers = 
                sr.getControllers();
        if(controllers == null || controllers.isEmpty()){
            return;
        }
        for(ServoController controller : controllers.values()){
            context.registerService(ServoController.class.getName(), 
                    controller, new Properties());
        }
    }
    
    private File chooseFile(){
        JFileChooser chooser = new JFileChooser();
        int ret = chooser.showOpenDialog(null);
        if(ret != JFileChooser.APPROVE_OPTION){
            return null;
        }
        return chooser.getSelectedFile();
    }
    
    private File predfinedFile(String path){
        return ConfigUtils.getFileSystemAdapter().openFile(path);
    }
    
    private void loadJointGroup(BundleContext context){
        File file = ConfigUtils.getFileSystemAdapter().openFile("./resources/jointgroup.xml");
        if(file == null){
            return;
        }
        JointGroup group = ServiceConnectionDirectory.buildService(context, 
                RobotJointGroup.VERSION, 
                RobotJointGroupConfigXMLReader.VERSION, 
                file, File.class, JointGroup.class);
        if(group == null){
            return;
        }
        context.registerService(JointGroup.class.getName(), group, new Properties());
        theLogger.log(Level.INFO, "JointGroup Service Registered.");
    }
}
